MODULE field_mod
  USE genmod
  IMPLICIT NONE

  INTEGER,PARAMETER :: field_T=1
  INTEGER,PARAMETER :: field_U=2
  INTEGER,PARAMETER :: field_Z=3

  INTEGER,PARAMETER :: type_integer=1
  INTEGER,PARAMETER :: type_real=2
  INTEGER,PARAMETER :: type_logical=3

  {%- set ranks = ( (2,':'), (3,':,:'), (4,':,:,:') ) %}
  {%- set types = ( ('r','real','REAL(rstd)'), ('i','integer','INTEGER'), ('l','logical','LOGICAL') ) %}
  
  TYPE t_field
     CHARACTER(30)      :: name
     LOGICAL :: ondevice !< flag if field is allocated on device as well
     INTEGER :: ndim
     INTEGER :: field_type
     INTEGER :: data_type
     INTEGER :: dim3
     INTEGER :: dim4

     {%- for tn, tln, tp in types %} {%- for rk, shp in ranks %}     
     {{tp}}, POINTER, CONTIGUOUS :: {{tn}}val{{rk}}d({{shp}}) => NULL()
     {%- endfor %} {%- endfor %}
  END TYPE t_field

  INTERFACE get_val
     {%- for tn, tln, tp in types %} {%- for rk, shp in ranks %}     
     MODULE PROCEDURE getval_{{tn}}{{rk}}d
     {%- endfor %} {%- endfor %}
  END INTERFACE

  INTERFACE ASSIGNMENT(=)
     {%- for tn, tln, tp in types %} {%- for rk, shp in ranks %}     
     MODULE PROCEDURE getval_{{tn}}{{rk}}d
     {%- endfor %} {%- endfor %}
  END INTERFACE
  
  PRIVATE :: allocate_field_, deallocate_field_

CONTAINS

  !====================================== allocate_field ===================================

  SUBROUTINE allocate_field_glo(field,field_type,data_type,dim1,dim2,name)
    USE domain_mod
    TYPE(t_field),POINTER :: field(:)
    INTEGER,INTENT(IN) :: field_type
    INTEGER,INTENT(IN) :: data_type
    INTEGER,OPTIONAL :: dim1,dim2
    CHARACTER(*), OPTIONAL :: name
    INTEGER :: ind

    ALLOCATE(field(ndomain_glo))
    DO ind=1,ndomain_glo
       CALL allocate_field_(domain_glo(ind), field(ind), field_type, data_type, dim1, dim2, name)
    ENDDO

  END SUBROUTINE allocate_field_glo

  SUBROUTINE allocate_field(field, field_type, data_type, dim3, dim4, name, ondevice)
    USE domain_mod
    USE omp_para
    TYPE(t_field), POINTER :: field(:)
    INTEGER, INTENT(IN)    :: field_type
    INTEGER, INTENT(IN)    :: data_type
    INTEGER, OPTIONAL      :: dim3,dim4
    CHARACTER(*), OPTIONAL :: name
    LOGICAL, INTENT(IN), OPTIONAL :: ondevice
    INTEGER :: ind
    !$OMP BARRIER
    !$OMP MASTER
    ALLOCATE(field(ndomain))
    !$OMP END MASTER
    !$OMP BARRIER

    DO ind=1,ndomain
       IF (.NOT. assigned_domain(ind)  .OR. .NOT. is_omp_level_master) CYCLE
       CALL allocate_field_(domain(ind), field(ind), field_type, data_type, dim3, dim4, name, ondevice)
    END DO
    !$OMP BARRIER

  END SUBROUTINE allocate_field

  SUBROUTINE allocate_fields(nfield,field,field_type,data_type,dim3,dim4,name, ondevice)
    USE domain_mod
    USE omp_para
    INTEGER, INTENT(IN)     :: nfield
    TYPE(t_field), POINTER  :: field(:,:)
    INTEGER, INTENT(IN)     :: field_type
    INTEGER, INTENT(IN)     :: data_type
    INTEGER, OPTIONAL       :: dim3,dim4
    CHARACTER(*), OPTIONAL  :: name
    LOGICAL, INTENT(IN), OPTIONAL :: ondevice
    INTEGER :: i, ind
    !$OMP BARRIER
    !$OMP MASTER
    ALLOCATE(field(ndomain,nfield))
    !$OMP END MASTER
    !$OMP BARRIER
    DO ind=1,ndomain
       IF (.NOT. assigned_domain(ind)  .OR. .NOT. is_omp_level_master) CYCLE
       DO i=1,nfield
          CALL allocate_field_(domain(ind), field(ind,i),field_type, data_type, dim3, dim4, name, ondevice)
       END DO
    END DO
    !$OMP BARRIER

  END SUBROUTINE allocate_fields

  SUBROUTINE allocate_field_(dom, field, field_type, data_type, dim3, dim4, name, ondevice)
    USE domain_mod
    USE omp_para
    TYPE(t_domain)         :: dom
    TYPE(t_field)          :: field
    INTEGER, INTENT(IN)    :: field_type
    INTEGER, INTENT(IN)    :: data_type
    INTEGER, OPTIONAL      :: dim3,dim4
    CHARACTER(*), OPTIONAL :: name
    LOGICAL, INTENT(IN), OPTIONAL :: ondevice

    INTEGER :: ij_size

    IF(PRESENT(name)) THEN
       field%name = name
    ELSE
       field%name = '(undefined)'
    END IF

    IF (PRESENT(dim4)) THEN
       field%ndim=4
       field%dim4=dim4
       field%dim3=dim3
    ELSE IF (PRESENT(dim3)) THEN
       field%ndim=3
       field%dim3=dim3
       field%dim4=1
    ELSE
       field%ndim=2
       field%dim3=1
       field%dim4=1
    ENDIF


    field%data_type=data_type
    field%field_type=field_type

    IF (field_type==field_T) THEN
       ij_size=dom%iim*dom%jjm
    ELSE IF (field_type==field_U) THEN
       ij_size=3*dom%iim*dom%jjm
    ELSE IF (field_type==field_Z) THEN
       ij_size=2*dom%iim*dom%jjm
    ENDIF

    IF (data_type==type_integer) ALLOCATE(field%ival4d(ij_size, field%dim3, field%dim4))
    IF (data_type==type_real)    ALLOCATE(field%rval4d(ij_size, field%dim3, field%dim4))
    IF (data_type==type_logical) ALLOCATE(field%lval4d(ij_size, field%dim3, field%dim4))

    IF (field%ndim==3) THEN
       IF (data_type==type_integer) field%ival3d => field%ival4d(:,:,1)
       IF (data_type==type_real)    field%rval3d => field%rval4d(:,:,1)
       IF (data_type==type_logical) field%lval3d => field%lval4d(:,:,1)

    ELSE IF (field%ndim==2) THEN
       IF (data_type==type_integer) field%ival2d => field%ival4d(:,1,1)
       IF (data_type==type_real)    field%rval2d => field%rval4d(:,1,1)
       IF (data_type==type_logical) field%lval2d => field%lval4d(:,1,1)

    ENDIF

    field%ondevice = .FALSE.
    IF (PRESENT(ondevice)) THEN
       IF (ondevice) CALL create_device_field(field)
    END IF

  END SUBROUTINE allocate_field_

  !==================================== deallocate_field ===================================

  SUBROUTINE deallocate_field_glo(field)
    USE domain_mod
    TYPE(t_field),POINTER :: field(:)
    INTEGER :: ind
    DO ind=1,ndomain_glo
       CALL deallocate_field_(field(ind))
    END DO
    DEALLOCATE(field)
  END SUBROUTINE deallocate_field_glo

  SUBROUTINE deallocate_field(field)
    USE domain_mod
    USE omp_para
    TYPE(t_field),POINTER :: field(:)
    INTEGER :: ind
    !$OMP BARRIER
    DO ind=1,ndomain
       IF (.NOT. assigned_domain(ind)  .OR. .NOT. is_omp_level_master) CYCLE
       CALL deallocate_field_(field(ind))
    END DO
    !$OMP BARRIER
    !$OMP MASTER
    DEALLOCATE(field)
    !$OMP END MASTER
    !$OMP BARRIER
  END SUBROUTINE deallocate_field

  SUBROUTINE deallocate_fields(field)
    USE domain_mod
    USE omp_para
    TYPE(t_field),POINTER :: field(:,:)
    INTEGER :: i, ind
    !$OMP BARRIER
    DO ind=1,ndomain
       IF (.NOT. assigned_domain(ind)  .OR. .NOT. is_omp_level_master) CYCLE
       DO i=1,SIZE(field,2)
          CALL deallocate_field_(field(ind,i))
       END DO
    END DO
    !$OMP BARRIER
    !$OMP MASTER
    DEALLOCATE(field)
    !$OMP END MASTER
    !$OMP BARRIER
  END SUBROUTINE deallocate_fields

  SUBROUTINE deallocate_field_(field)
    USE domain_mod
    USE omp_para
    TYPE(t_field) :: field
    INTEGER :: data_type
    data_type=field%data_type

    {%- for tn, tln, tp in types %}
    IF (data_type==type_{{tln}}) THEN
       IF (field%ondevice) THEN
          !$acc exit data delete(field%{{tn}}val4d(:,:,:))
          CONTINUE
       END IF
       DEALLOCATE(field%{{tn}}val4d)
    END IF
    {%- endfor %}

  END SUBROUTINE deallocate_field_

  !====================================== getval ===================================

  {%- for tn, tln, tp in types %} {%- for rk, shp in ranks %}

  SUBROUTINE getval_{{tn}}{{rk}}d(field_pt,field)
    {{tp}}, POINTER, INTENT(INOUT) :: field_pt({{shp}})
    TYPE(t_field),INTENT(IN) :: field

    IF (field%ndim/={{rk}} .OR. field%data_type/=type_{{tln}}) THEN
       PRINT *, 'getval_{{tn}}{{rk}}d : bad pointer assignment with ' // TRIM(field%name)
       STOP
    END IF
    field_pt=>field%{{tn}}val{{rk}}d
  END SUBROUTINE getval_{{tn}}{{rk}}d

  {%- endfor %} {%- endfor %}

  !===================== Data transfer between host (CPU) and device (GPU) =========================

  SUBROUTINE update_device_field(field)
    USE domain_mod
    USE omp_para
    TYPE(t_field) :: field(:)
    INTEGER :: ind

    DO ind=1,ndomain
       IF (.NOT. assigned_domain(ind)  .OR. .NOT. is_omp_level_master) CYCLE

       IF (.NOT. field(ind)%ondevice) CALL create_device_field(field(ind))

       {%- for tn, tln, tp in types %}
       IF (field(ind)%data_type==type_{{tln}}) THEN
          !$acc update device(field(ind)%{{tn}}val4d(:,:,:)) async
          CONTINUE
       END IF
       {%- endfor %}

    ENDDO
    !$OMP BARRIER
  END SUBROUTINE update_device_field

  SUBROUTINE update_host_field(field)
    USE domain_mod
    USE omp_para
    TYPE(t_field) :: field(:)
    INTEGER :: ind

    DO ind=1,ndomain
       IF (.NOT. assigned_domain(ind)  .OR. .NOT. is_omp_level_master) CYCLE

       IF (field(ind)%ondevice) THEN
          
          {%- for tn, tln, tp in types %}
          IF (field(ind)%data_type==type_{{tln}}) THEN
             !$acc update host(field(ind)%{{tn}}val4d(:,:,:)) async
             CONTINUE
          END IF
          {%- endfor %}
          
       END IF
    ENDDO
    !$acc wait
    !$OMP BARRIER
  END SUBROUTINE update_host_field

  SUBROUTINE create_device_field(field)
    TYPE(t_field) :: field

    IF (field%ondevice) THEN
       PRINT *, "Field is already on device !"
       STOP 1
    END IF
    
    {%- for tn, tln, tp in types %}
    IF (field%data_type==type_{{tln}}) THEN
       !$acc enter data create(field%{{tn}}val4d(:,:,:)) async
    END IF
    {%- endfor %}
    
    field%ondevice = .TRUE.
  END SUBROUTINE create_device_field

END MODULE field_mod
